# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
from .....schema_classes import AccessLevelClass
from .....schema_classes import AuditStampClass
from .....schema_classes import BrowsePathsClass
from .....schema_classes import ChangeAuditStampsClass
from .....schema_classes import CostClass
from .....schema_classes import CostCostClass
from .....schema_classes import CostCostDiscriminatorClass
from .....schema_classes import CostTypeClass
from .....schema_classes import DataPlatformInstanceClass
from .....schema_classes import DeprecationClass
from .....schema_classes import FabricTypeClass
from .....schema_classes import GlobalTagsClass
from .....schema_classes import GlossaryTermAssociationClass
from .....schema_classes import GlossaryTermsClass
from .....schema_classes import InputFieldClass
from .....schema_classes import InputFieldsClass
from .....schema_classes import InstitutionalMemoryClass
from .....schema_classes import InstitutionalMemoryMetadataClass
from .....schema_classes import MLFeatureDataTypeClass
from .....schema_classes import MediaClass
from .....schema_classes import MediaTypeClass
from .....schema_classes import OperationClass
from .....schema_classes import OperationSourceTypeClass
from .....schema_classes import OperationTypeClass
from .....schema_classes import OriginClass
from .....schema_classes import OriginTypeClass
from .....schema_classes import OwnerClass
from .....schema_classes import OwnershipClass
from .....schema_classes import OwnershipSourceClass
from .....schema_classes import OwnershipSourceTypeClass
from .....schema_classes import OwnershipTypeClass
from .....schema_classes import SiblingsClass
from .....schema_classes import StatusClass
from .....schema_classes import SubTypesClass
from .....schema_classes import TagAssociationClass
from .....schema_classes import TimeStampClass
from .....schema_classes import VersionTagClass
from .....schema_classes import WindowDurationClass


AccessLevel = AccessLevelClass
AuditStamp = AuditStampClass
BrowsePaths = BrowsePathsClass
ChangeAuditStamps = ChangeAuditStampsClass
Cost = CostClass
CostCost = CostCostClass
CostCostDiscriminator = CostCostDiscriminatorClass
CostType = CostTypeClass
DataPlatformInstance = DataPlatformInstanceClass
Deprecation = DeprecationClass
FabricType = FabricTypeClass
GlobalTags = GlobalTagsClass
GlossaryTermAssociation = GlossaryTermAssociationClass
GlossaryTerms = GlossaryTermsClass
InputField = InputFieldClass
InputFields = InputFieldsClass
InstitutionalMemory = InstitutionalMemoryClass
InstitutionalMemoryMetadata = InstitutionalMemoryMetadataClass
MLFeatureDataType = MLFeatureDataTypeClass
Media = MediaClass
MediaType = MediaTypeClass
Operation = OperationClass
OperationSourceType = OperationSourceTypeClass
OperationType = OperationTypeClass
Origin = OriginClass
OriginType = OriginTypeClass
Owner = OwnerClass
Ownership = OwnershipClass
OwnershipSource = OwnershipSourceClass
OwnershipSourceType = OwnershipSourceTypeClass
OwnershipType = OwnershipTypeClass
Siblings = SiblingsClass
Status = StatusClass
SubTypes = SubTypesClass
TagAssociation = TagAssociationClass
TimeStamp = TimeStampClass
VersionTag = VersionTagClass
WindowDuration = WindowDurationClass

# fmt: on
